<?php
/**
 * Enumeration of all the different GDF data types.
 */
final class DATATYPE {
    const INT = 0;
    const LONG = 1;
    const FLOAT = 2;
    const DOUBLE = 3;
    const BOOLEAN = 4;
    const STRING = 5;
    const Int = 0;
    const Long = 1;
    const Float = 2;
    const Double = 3;
    const Boolean = 4;
    const String = 5;

    /**
     * Get length, in bytes, required to store a fixed length value of the given type.  Returns -1 for Strings.
     * @param DATATYPE $type The datatype to get the length of
     * @return int The length in bytes
     */
    static function getLength($type) {
        switch($type) {
            case DATATYPE::INT:
            case DATATYPE::FLOAT:
                return 4;
            case DATATYPE::LONG:
            case DATATYPE::DOUBLE:
                return 8;
            case DATATYPE::BOOLEAN:
                return 1;
            default:
                return -1;
        }
    }
    
    static function toString($type) {
        $temp = Array("INT", "LONG", "FLOAT", "DOUBLE", "BOOLEAN", "STRING");
        return $temp[$type];
    }
}

/**
 * A single GDF entry, representing a type and a value.
 */
class Entry {

    var $datatype;
    var $value;
        
    /**
     * Creates a new Entry with the given type and value
     * @param DATATYPE $datatype
     * @param mixed $value
     */
    function __construct($datatype, $value) {
        $this->datatype = $datatype;
        $this->value = $value;
    }
        
    /**
     * Gets the storable representation of the entry's value in bytes.
     * @return byte[] The bytes of the value
     */
    function getBytes() {
        $length = 0;
        switch ($this->datatype) {
            default:
                $length = DATATYPE::getLength($this->datatype);
                break;
            case DATATYPE::STRING:
                $length = strlen($this->value) + 4;
                break;
        }
        $ret = Array(length + 1);
        $ret[0] = $this->datatype;
        switch (datatype) {
            case INT:
                $ret[1] = $this->value >> 24 & 0xFF;
                $ret[2] = $this->value >> 16 & 0xFF;
                $ret[3] = $this->value >> 8 & 0xFF;
                $ret[4] = $this->value & 0xFF;
                break;
            case FLOAT:
                $ret[1] = $this->value >> 24 & 0xFF;
                $ret[2] = $this->value >> 16 & 0xFF;
                $ret[3] = $this->value >> 8 & 0xFF;
                $ret[4] = $this->value & 0xFF;
                break;
            case LONG:
                $ret[1] = ($this->value >> 56 & 0xFF);
                $ret[2] = ($this->value >> 48 & 0xFF);
                $ret[3] = ($this->value >> 40 & 0xFF);
                $ret[4] = ($this->value >> 32 & 0xFF);
                $ret[5] = ($this->value >> 24 & 0xFF);
                $ret[6] = ($this->value >> 16 & 0xFF);
                $ret[7] = ($this->value >> 8 & 0xFF);
                $ret[8] = ($this->value & 0xFF);
                break;
            case DOUBLE:
                $ret[1] = ($this->value >> 56 & 0xFF);
                $ret[2] = ($this->value >> 48 & 0xFF);
                $ret[3] = ($this->value >> 40 & 0xFF);
                $ret[4] = ($this->value >> 32 & 0xFF);
                $ret[5] = ($this->value >> 24 & 0xFF);
                $ret[6] = ($this->value >> 16 & 0xFF);
                $ret[7] = ($this->value >> 8 & 0xFF);
                $ret[8] = ($this->value & 0xFF);
                break;
            case BOOLEAN:
                $ret[1] = $this->value ? 1 : 0;
                break;
            case STRING:
                $ret[1] = ($length - 4) >> 24 & 0xFF;
                $ret[2] = ($length - 4) >> 16 & 0xFF;
                $ret[3] = ($length - 4) >> 8 & 0xFF;
                $ret[4] = ($length - 4) & 0xFF;
                for ($x = 0; $x < strlen($this->value); $x++) {
                    $ret[5 + x] = $this->value[$x];
                }
                break;
        }
        return $ret;
    }
        
    /**
     * QQQ
     * @param byte[] $bytes
     * @return Entry
     */
    static function getEntryFromBytes($bytes) {
        $value = null;
        $datatype = null;
        switch ($bytes[0]) {
            case 0: $datatype = DATATYPE::INT; break;
            case 1: $datatype = DATATYPE::LONG; break;
            case 2: $datatype = DATATYPE::FLOAT; break;
            case 3: $datatype = DATATYPE::DOUBLE; break;
            case 4: $datatype = DATATYPE::BOOLEAN; break;
            case 5: $datatype = DATATYPE::STRING; break;
        }
        switch ($datatype) {
            case DATATYPE::INT:
                $ivalue = 0;
                $ivalue += $bytes[1] << 24;
                $ivalue += $bytes[2] << 16;
                $ivalue += $bytes[3] << 8;
                $ivalue += $bytes[4];
                $value = $ivalue;
                break;
            case DATATYPE::FLOAT:
                $fvalue = 0;
                $fvalue += $bytes[1] << 24;
                $fvalue += $bytes[2] << 16;
                $fvalue += $bytes[3] << 8;
                $fvalue += $bytes[4] << 0;
                $value = unpack('f', pack('i', $fvalue));
                $value = $value[1];
                break;
            case DATATYPE::LONG:
                $lvalue = 0;
                $lvalue += $bytes[1] << 56;
                $lvalue += $bytes[2] << 48;
                $lvalue += $bytes[3] << 40;
                $lvalue += $bytes[4] << 32;
                $lvalue += $bytes[5] << 24;
                $lvalue += $bytes[6] << 16;
                $lvalue += $bytes[7] << 8;
                $lvalue += $bytes[8];
                $value  = $lvalue;
                break;
            case DATATYPE::DOUBLE:
                $dvalue1 = 0;
                $dvalue1 += $bytes[1] << 24;
                $dvalue1 += $bytes[2] << 16;
                $dvalue1 += $bytes[3] << 8;
                $dvalue1 += $bytes[4] << 0;
                $dvalue2 = 0;
                $dvalue2 += $bytes[5] << 24;
                $dvalue2 += $bytes[6] << 16;
                $dvalue2 += $bytes[7] << 8;
                $dvalue2 += $bytes[8] << 0;
                $value = unpack('d', pack('i', $dvalue2) . pack('i', $dvalue1));
                $value = $value[1];
                break;
            case DATATYPE::BOOLEAN:
                $value = $bytes[1];
                break;
            case DATATYPE::STRING:
                $length = 0;
                $svalue = "";
                $length += $bytes[1] << 24;
                $length += $bytes[2] << 16;
                $length += $bytes[3] << 8;
                $length += $bytes[4];
                for ($x = 0; $x < $length; $x++) {
                    $svalue .= chr($bytes[5 + $x]);
                }
                $value = $svalue;
                break;
        }
        return new Entry($datatype, $value);
    }
}

/**
 * <b>Global Data Format</b><br>
 * A simple yet powerful way to save simple data, with a similar functionality to Java serialization, but more "global" as it is not java specific, and relies simply on a stack of different values of "basic" data types.
 * @author Ricky Talbot
 */
class GDF {
    private $data = Array();
    
    /**
     * Optionally initializes the GDF with a starting stack of entries.
     * @param Entry[] data The initial entry stack
     */
    function __construct($data = Array()) {
        $this->data = $data;
    }
    
    /**
     * Pushes an entry onto the top of the entry stack.
     * @param Entry $entry The entry to push
     */
    function pushEntry($entry) {
        array_push($this->data, $entry);
    }
    
    /**
     * Pops an entry off of the top of the entry stack and returns it.
     * @return Entry The popped entry
     */
    function popEntry() {
        return array_pop($this->data);
    }
    
    /**
     * Pops an entry off of the bottom of the entry stack and returns it.
     * @return Entry The popped entry
     */
    function rpopEntry() {
        $temp = array_splice($this->data, 0, 1);
        return $temp[0];
    }
    
    /**
     * Gets and returns an entry without removing it at the specified index in the entry stack.
     * @param int $index Index of the entry
     * @return Entry The entry at the specified index
     */
    function getEntry($index) {
        return $this->data[$index];
    }
    
    /**
     * Gets the total size of the data stack in entries.
     * @return int The stack size
     */
    function getSize() {
        return count($this->data);
    }
    
    /**
     * Tests if the stack has no entries.
     * @return boolean <var>true</var> if the stack has zero elements, <var>false</var> otherwise
     */
    function isEmtpty() {
        return count($this->data) == 0;
    }
    
    /**
     * 
     * @param resource $handle
     */
    function writeGDF($handle) {
        for ($x = 0; $x < data.size(); $x++) {
            $bytes = $this->data[$x]->getBytes();
            for ($y = 0; $y < $bytes.length; $y++) {
                fwrite($handle, chr($bytes[$y]));
                writeByte($handle, $bytes[$y]);
            }
        }
    }
    
    /**
     * 
     * @param resource $handle
     * @throws IOException
     */
    function readGDF($handle) {
        $bytes = Array();
        $temp;
        $count = 0;
//        while ($count < filesize("./test.gdf")) {
//            $temp = fread($handle, 1);
//            array_push($bytes, ord($temp));
//            $count++;
//        }
        $temp = fread($handle, filesize("./test.gdf"));
        for ($x = 0; $x < strlen($temp); $x++) $bytes[$x] = ord($temp[$x]);
        $this->data = Array();
        $x = 0;
        while ($x < count($bytes)) {
            $datatype = -1;
            switch ($bytes[$x]) {
                case 0: $datatype = DATATYPE::INT; break;
                case 1: $datatype = DATATYPE::LONG; break;
                case 2: $datatype = DATATYPE::FLOAT; break;
                case 3: $datatype = DATATYPE::DOUBLE; break;
                case 4: $datatype = DATATYPE::BOOLEAN; break;
                case 5: $datatype = DATATYPE::STRING; break;
            }
            $length = -1;
            switch ($datatype) {
                default:
                    $length = DATATYPE::getLength($datatype);
                    break;
                case DATATYPE::STRING:
                    $length = 4;
                    $length += $bytes[$x + 1] << 24;
                    $length += $bytes[$x + 2] << 16;
                    $length += $bytes[$x + 3] << 8;
                    $length += $bytes[$x + 4];
                    break;
            }
            $length += 1;
            $entrybytes = Array();
            if ($length > 1000) break;
            for ($y = 0; $y < $length; $y++) {
                array_push($entrybytes, $bytes[$x + $y]);
            }
            array_push($this->data, Entry::getEntryFromBytes($entrybytes));
            $x += $length;
        }
    }
}
?>