import java.io.*;
import java.util.*;

/**
 * <b>Global Data Format</b><br>
 * A simple yet powerful way to save simple data, with a similar functionality to Java serialization, but more "global" as it is not java specific, and relies simply on a stack of different values of "basic" data types.
 * @author Ricky Talbot
 */
public class GDF implements Cloneable {
        
    /**
     * Enumeration of all the different GDF data types.
     */
    public static enum DATATYPE {

        /**
         * 
         */
        INT(0),
        /**
         * 
         */
        LONG(1),
        /**
         * 
         */
        FLOAT(2),
        /**
         * 
         */
        DOUBLE(3),
        /**
         * 
         */
        BOOLEAN(4),
        /**
         * 
         */
        STRING(5),
        /**
         * 
         */
        Int(0),
        /**
         * 
         */
        Long(1),
        /**
         * 
         */
        Float(2),
        /**
         * 
         */
        Double(3),
        /**
         * 
         */
        Boolean(4),
        /**
         * 
         */
        String(5);
        
        private int refnum;
        
        private DATATYPE(int refnum) {
            this.refnum = refnum;
        }
        
        /**
         * Gets the unique GDF identifier representing which type of data this is numerically.
         * @return
         */
        public int getRefNum() {
            return refnum;
        }
        
        /**
         * Get length, in bytes, required to store a fixed length value of this type.  Returns -1 for Strings.
         * @return The length in bytes
         */
        public int getLength() {
            switch (this) {
                case INT:
                case Int:
                case FLOAT:
                case Float:
                    return 4;
                case LONG:
                case Long:
                case DOUBLE:
                case Double:
                    return 8;
                case BOOLEAN:
                case Boolean:
                    return 1;
                default:
                    return -1;
            }
        }
    }
    
    /**
     * A single GDF entry, representing a type and a value.
     */
    public static class Entry {
        
        public DATATYPE datatype;
        public Object value;
        public Integer ivalue;
        public Float fvalue;
        public Long lvalue;
        public Double dvalue;
        public Boolean bvalue;
        public String svalue;
        
        /**
         * Creates a new Entry with the given type and value
         * @param datatype
         * @param value
         */
        public Entry(DATATYPE datatype, Object value) {
            this.datatype = datatype;
            this.value = value;
            if (value instanceof Integer) {
                ivalue = (Integer)value;
            } else if (value instanceof Short) {
                ivalue = ((Short)value).intValue();
            } else if (value instanceof Long) {
                lvalue = (Long)value;
            } else if (value instanceof Double) {
                dvalue = (Double)value;
            } else if (value instanceof Float) {
                fvalue = (Float)value;
            } else if (value instanceof Boolean) {
                bvalue = (Boolean)value;
            } else if (value instanceof String) {
                svalue = (String)value;
            } 
        }
        
        /**
         * Gets the storable representation of the entry's value in bytes.
         * @return The bytes of the value
         */
        public int[] getBytes() {
            int length = 0;
            switch (datatype) {
                default:
                    length = datatype.getLength();
                    break;
                case STRING:
                case String:
                    length = svalue.length() + 4;
                    break;
            }
            int[] ret = new int[length + 1];
            ret[0] = datatype.getRefNum();
            int tempivalue;
            long templvalue;
            switch (datatype) {
                case INT:
                case Int:
                    ret[1] = ivalue >> 24 & 0xFF;
                    ret[2] = ivalue >> 16 & 0xFF;
                    ret[3] = ivalue >> 8 & 0xFF;
                    ret[4] = ivalue & 0xFF;
                    break;
                case FLOAT:
                case Float:
                    tempivalue = Float.floatToRawIntBits(fvalue);
                    ret[1] = tempivalue >> 24 & 0xFF;
                    ret[2] = tempivalue >> 16 & 0xFF;
                    ret[3] = tempivalue >> 8 & 0xFF;
                    ret[4] = tempivalue & 0xFF;
                    break;
                case LONG:
                case Long:
                    ret[1] = (int)(lvalue >> 56 & 0xFF);
                    ret[2] = (int)(lvalue >> 48 & 0xFF);
                    ret[3] = (int)(lvalue >> 40 & 0xFF);
                    ret[4] = (int)(lvalue >> 32 & 0xFF);
                    ret[5] = (int)(lvalue >> 24 & 0xFF);
                    ret[6] = (int)(lvalue >> 16 & 0xFF);
                    ret[7] = (int)(lvalue >> 8 & 0xFF);
                    ret[8] = (int)(lvalue & 0xFF);
                    break;
                case DOUBLE:
                case Double:
                    templvalue = Double.doubleToRawLongBits(dvalue);
                    ret[1] = (int)(templvalue >> 56 & 0xFF);
                    ret[2] = (int)(templvalue >> 48 & 0xFF);
                    ret[3] = (int)(templvalue >> 40 & 0xFF);
                    ret[4] = (int)(templvalue >> 32 & 0xFF);
                    ret[5] = (int)(templvalue >> 24 & 0xFF);
                    ret[6] = (int)(templvalue >> 16 & 0xFF);
                    ret[7] = (int)(templvalue >> 8 & 0xFF);
                    ret[8] = (int)(templvalue & 0xFF);
                    break;
                case BOOLEAN:
                case Boolean:
                    ret[1] = bvalue ? 1 : 0;
                    break;
                case STRING:
                case String:
                    ret[1] = (length - 4) >> 24 & 0xFF;
                    ret[2] = (length - 4) >> 16 & 0xFF;
                    ret[3] = (length - 4) >> 8 & 0xFF;
                    ret[4] = (length - 4) & 0xFF;
                    for (int x = 0; x < svalue.length(); x++) {
                        ret[5 + x] = svalue.charAt(x);
                    }
                    break;
            }
            return ret;
        }
        
        /**
         * QQQ
         * @param bytes
         * @return
         */
        public static Entry getEntryFromBytes(int[] bytes) {
            Object value = null;
            DATATYPE datatype = null;
            switch (bytes[0]) {
                case 0: datatype = DATATYPE.INT; break;
                case 1: datatype = DATATYPE.LONG; break;
                case 2: datatype = DATATYPE.FLOAT; break;
                case 3: datatype = DATATYPE.DOUBLE; break;
                case 4: datatype = DATATYPE.BOOLEAN; break;
                case 5: datatype = DATATYPE.STRING; break;
            }
            switch (datatype) {
                case INT:
                    int ivalue = 0;
                    ivalue += bytes[1] << 24;
                    ivalue += bytes[2] << 16;
                    ivalue += bytes[3] << 8;
                    ivalue += bytes[4];
                    value = ivalue;
                    break;
                case FLOAT:
                    int fvalue = 0;
                    fvalue += bytes[1] << 24;
                    fvalue += bytes[2] << 16;
                    fvalue += bytes[3] << 8;
                    fvalue += bytes[4];
                    value = Float.intBitsToFloat(fvalue);
                    break;
                case LONG:
                    long lvalue = 0;
                    lvalue += (long)bytes[1] << 56;
                    lvalue += (long)bytes[2] << 48;
                    lvalue += (long)bytes[3] << 40;
                    lvalue += (long)bytes[4] << 32;
                    lvalue += (long)bytes[5] << 24;
                    lvalue += (long)bytes[6] << 16;
                    lvalue += (long)bytes[7] << 8;
                    lvalue += (long)bytes[8];
                    value = lvalue;
                    break;
                case DOUBLE:
                    long dvalue = 0;
                    dvalue += (long)bytes[1] << 56;
                    dvalue += (long)bytes[2] << 48;
                    dvalue += (long)bytes[3] << 40;
                    dvalue += (long)bytes[4] << 32;
                    dvalue += (long)bytes[5] << 24;
                    dvalue += (long)bytes[6] << 16;
                    dvalue += (long)bytes[7] << 8;
                    dvalue += (long)bytes[8];
                    value = Double.longBitsToDouble(dvalue);
                    break;
                case BOOLEAN:
                    value = (bytes[1] == 1);
                    break;
                case STRING:
                    int length = 0;
                    String svalue = "";
                    length += bytes[1] << 24;
                    length += bytes[2] << 16;
                    length += bytes[3] << 8;
                    length += bytes[4];
                    for (int x = 0; x < length; x++) {
                        svalue += (char)bytes[5 + x];
                    }
                    value = svalue;
                    break;
            }
            return new Entry(datatype, value);
        }
    }
    
    private Stack<Entry> data;
    
    /**
     * 
     */
    public GDF() {
        this.data = new Stack<Entry>();
    }
    
    /**
     * Initializes the GDF with a starting stack of entries.
     * @param dataThe initial entry stack
     */
    public GDF(Stack<Entry> data) {
        this.data = data;
    }
    
    /**
     * Pushes an entry onto the top of the entry stack.
     * @param entry The entry to push
     */
    public void pushEntry(Entry entry) {
        data.push(entry);
    }
    
    /**
     * Pops an entry off of the top of the entry stack and returns it.
     * @return The popped entry
     */
    public Entry popEntry() {
        return data.pop();
    }
    
    /**
     * Pops an entry off of the bottom of the entry stack and returns it.
     * @return The popped entry
     */
    public Entry rpopEntry() {
        return data.remove(0);
    }
    
    /**
     * Gets and returns an entry without removing it at the specified index in the entry stack.
     * @param index Index of the entry
     * @return The entry at the specified index
     */
    public Entry getEntry(int index) {
        return data.get(index);
    }
    
    /**
     * Gets the total size of the data stack in entries.
     * @return The stack size
     */
    public int getSize() {
        return data.size();
    }
    
    /**
     * Tests if the stack has no entries.
     * @return <tt>true</tt> if the stack has zero elements, <tt>false</tt> otherwise
     */
    public boolean isEmtpty() {
        return data.isEmpty();
    }
    
    /**
     * 
     * @param os
     * @throws IOException
     */
    public void writeGDF(OutputStream os) throws IOException {
        for (int x = 0; x < data.size(); x++) {
            int[] bytes = data.get(x).getBytes();
            for (int y = 0; y < bytes.length; y++) {
                os.write(bytes[y]);
            }
        }
    }
    
    /**
     * 
     * @param is
     * @throws IOException
     */
    public void readGDF(InputStream is) throws IOException {
        Stack<Integer> i = new Stack<Integer>();
        int temp;
        while ((temp = is.read()) != -1) {
            i.push(temp);
        }
        Integer[] Ibytes = i.toArray(new Integer[0]);
        int[] bytes = new int[Ibytes.length];
        for (int x = 0; x < Ibytes.length; x++) {
            bytes[x] = Ibytes[x];
        }
        data = new Stack<Entry>();
        int x = 0;
        while (x < bytes.length) {
            DATATYPE datatype = null;
            switch (bytes[x]) {
                case 0: datatype = DATATYPE.INT; break;
                case 1: datatype = DATATYPE.LONG; break;
                case 2: datatype = DATATYPE.FLOAT; break;
                case 3: datatype = DATATYPE.DOUBLE; break;
                case 4: datatype = DATATYPE.BOOLEAN; break;
                case 5: datatype = DATATYPE.STRING; break;
            }
            int length = -1;
            switch (datatype) {
                default:
                    length = datatype.getLength();
                    break;
                case STRING:
                    length = 4;
                    length += bytes[x + 1] << 24;
                    length += bytes[x + 2] << 16;
                    length += bytes[x + 3] << 8;
                    length += bytes[x + 4];
                    break;
            }
            length += 1;
            int[] entrybytes = new int[length];
            System.arraycopy(bytes, x, entrybytes, 0, length);
            data.push(Entry.getEntryFromBytes(entrybytes));
            x += length;
        }
    }
    
    public GDF clone() {
        return new GDF((Stack<Entry>)data.clone());
    }
}
